# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import
"""
Test for version command of virt-who

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import os
import sys

from base import TestBase
from virtwho.parser import get_version

class TestVersion(TestBase):

    def test_version_value(self):
        version = get_version()
        self.assertTrue(version[:9] == "virt-who ")
        self.assertGreater(len(version), 14)
